# -*- coding: utf-8 -*-
"""
Created on Thu May 20 17:06:19 2021

@author: Dang Trung Anh
"""

from sklearn import preprocessing
from pathlib import Path

from .optimizers import CSSA as cssa
from .optimizers import CPSO as cpso
from .optimizers import CGA as cga
from .optimizers import CBAT as cbat
from .optimizers import CFFA as cffa
from .optimizers import CGWO as cgwo
from .optimizers import CWOA as cwoa
from .optimizers import CMVO as cmvo
from .optimizers import CMFO as cmfo
from .optimizers import CCS as ccs

from . import _objectives as objectives
from . import _measures as measures

import os
import sys
import numpy
import warnings
import time
import csv
from . import _plot_convergence as conv_plot
from . import _plot_boxplot as box_plot
from . import _cluster_detection as clus_det

warnings.simplefilter(action='ignore')


class EvoCluster:
    """ EvoCluster Library

    Parameters
    ----------
    optimizer: list
        The list of optimizers names
    objective_func: ...
    dataset_list: ...
    num_of_runs: ...
    params: ...
    export_flags: ...
    auto_cluster: ...
    n_clusters: ...
    labels_exist: ...
    metric: ...
    """

    def __init__(self, optimizer, objective_func, dataset_list, num_of_runs, params, export_flags,
                 auto_cluster=True, n_clusters='supervised', labels_exist=True, metric='euclidean'):

        self.optimizer = optimizer
        self.objective_func = objective_func
        self.dataset_list = dataset_list
        self.num_of_runs = num_of_runs
        self.params = params
        self.export_flags = export_flags
        self.auto_cluster = auto_cluster
        self.n_clusters = n_clusters
        self.labels_exist = labels_exist
        self.metric = metric

        # if not labels_exist and n_clusters == 'supervised':
        #     print(
        #         'Supervised value for n_clusters is not allowed when labels_exist value is false')
        #     sys.exit()

    def run(self):
        if not self.labels_exist and self.n_clusters == 'supervised':
            print(
                'Syupervised value for n_clusters is not allowed when labels_exist value is false')
            sys.exit()

        if isinstance(self.n_clusters, list):
            if len(self.n_clusters) != len(self.dataset_List):
                print(
                    'Length of n_clusters list should equal the length of dataset_List list')
                sys.exit()
            if min(self._clusters) < 2:
                print('n_clusters value should be larger than 2')
                sys.exit()
            if self.auto_cluster == True:
                print('n_clusters should be string if auto_cluster is true')
                sys.exit()
        else:
            if self.auto_cluster == False:
                print('n_clusters should be a list of integers if auto_cluster is false')
                sys.exit()

        # Select general parameters for all optimizers (population size, number of iterations) ....
        PopulationSize = self.params['PopulationSize']
        Iterations = self.params['Iterations']

        # Export results ?
        Export = self.export_flags['Export_avg']
        Export_details = self.export_flags['Export_details']
        Export_details_labels = self.export_flags['Export_details_labels']
        Export_convergence = self.export_flags['Export_convergence']
        Export_boxplot = self.export_flags['Export_boxplot']

        # Check if it works at least once
        Flag = False
        Flag_details = False
        Flag_details_Labels = False

        # CSV Header for for the cinvergence
        CnvgHeader = []

        if self.labels_exist:
            datasets_directory = "datasets/"  # the directory where the dataset is stored
        else:
            # the directory where the dataset is stored
            datasets_directory = "datasets/unsupervised/"

        results_directory = time.strftime("%Y-%m-%d-%H-%M-%S") + '/'
        Path(results_directory).mkdir(parents=True, exist_ok=True)

        dataset_len = len(self.dataset_list)

        k = [-1] * dataset_len
        f = [-1] * dataset_len
        points = [0] * dataset_len
        labelsTrue = [0] * dataset_len

        for l in range(0, Iterations):
            CnvgHeader.append("Iter"+str(l+1))

        # read all datasets
        for h in range(dataset_len):

            dataset_filename = self.dataset_list[h] + '.csv'
            # Read the dataset file and generate the points list and true values
            rawData = open(os.path.join(os.path.abspath(os.path.dirname(
                __file__)), datasets_directory + dataset_filename), 'rt')
            data = numpy.loadtxt(rawData, delimiter=",")

            nPoints, nValues = data.shape  # Number of points and Number of values for each point

            if self.labels_exist:
                f[h] = nValues - 1  # Dimension value
                points[h] = data[:, :-1].tolist()  # list of points
                # List of actual cluster of each points (last field)
                labelsTrue[h] = data[:, -1].tolist()
            else:
                f[h] = nValues  # Dimension value
                points[h] = data.copy().tolist()  # list of points
                # List of actual cluster of each points (last field)
                labelsTrue[h] = None

            points[h] = preprocessing.normalize(points[h], norm='max', axis=0)

            if self.n_clusters == 'supervised':
                k[h] = len(numpy.unique(data[:, -1]))  # k: Number of clusters
            elif self.n_clusters == 'elbow':
                k[h] = clus_det.ELBOW(points[h])  # k: Number of clusters
            elif self.n_clusters == 'gap':
                k[h] = clus_det.GAP_STATISTICS(
                    points[h])  # k: Number of clusters
            elif self.n_clusters == 'silhouette':
                k[h] = clus_det.SC(points[h])  # k: Number of clusters
            elif self.n_clusters == 'DB':
                k[h] = clus_det.DB(points[h])  # k: Number of clusters
            elif self.n_clusters == 'CH':
                k[h] = clus_det.CH(points[h])  # k: Number of clusters
            elif self.n_clusters == 'DB':
                k[h] = clus_det.DB(points[h])  # k: Number of clusters
            elif self.n_clusters == 'BIC':
                k[h] = clus_det.BIC(points[h])  # k: Number of clusters
            elif self.n_clusters == 'min':
                k[h] = clus_det.min_clusters(
                    points[h])  # k: Number of clusters
            elif self.n_clusters == 'max':
                k[h] = clus_det.max_clusters(
                    points[h])  # k: Number of clusters
            elif self.n_clusters == 'median':
                k[h] = clus_det.median_clusters(
                    points[h])  # k: Number of clusters
            elif self.n_clusters == 'majority':
                k[h] = clus_det.majority_clusters(
                    points[h])  # k: Number of clusters
            else:
                k[h] = self.n_clusters[h]  # k: Number of clusters

        for i in range(0, len(self.optimizer)):
            for j in range(0, len(self.objective_func)):
                for h in range(len(self.dataset_list)):
                    HS = [0]*self.num_of_runs
                    CS = [0]*self.num_of_runs
                    VM = [0]*self.num_of_runs
                    AMI = [0]*self.num_of_runs
                    ARI = [0]*self.num_of_runs
                    Fmeasure = [0]*self.num_of_runs
                    SC = [0]*self.num_of_runs
                    accuracy = [0]*self.num_of_runs
                    DI = [0]*self.num_of_runs
                    DB = [0]*self.num_of_runs
                    stdev = [0]*self.num_of_runs
                    exSSE = [0]*self.num_of_runs
                    exTWCV = [0]*self.num_of_runs
                    purity = [0]*self.num_of_runs
                    entropy = [0]*self.num_of_runs
                    convergence = [0]*self.num_of_runs
                    executionTime = [0]*self.num_of_runs
                    # Agg = [0]*NumOfRuns

                    for z in range(0, self.num_of_runs):
                        print("Dataset: " + self.dataset_list[h])
                        print("k: " + str(k[h]))
                        print("Run no.: " + str(z))
                        print("Population Size: " + str(PopulationSize))
                        print("Iterations: " + str(Iterations))

                        objective_name = self.objective_func[j]
                        x = self.selector(self.optimizer[i], objective_name, k[h], f[h],
                                    PopulationSize, Iterations, points[h], self.metric)

                        if self.labels_exist:
                            HS[z] = measures.HS(labelsTrue[h], x.labelsPred)
                            CS[z] = measures.CS(labelsTrue[h], x.labelsPred)
                            VM[z] = measures.VM(labelsTrue[h], x.labelsPred)
                            AMI[z] = measures.AMI(labelsTrue[h], x.labelsPred)
                            ARI[z] = measures.ARI(labelsTrue[h], x.labelsPred)
                            Fmeasure[z] = measures.Fmeasure(
                                labelsTrue[h], x.labelsPred)
                            accuracy[z] = measures.accuracy(
                                labelsTrue[h], x.labelsPred)
                            purity[z] = measures.purity(
                                labelsTrue[h], x.labelsPred)
                            entropy[z] = measures.entropy(
                                labelsTrue[h], x.labelsPred)
                            # Agg[z] = float("%0.2f"%(float("%0.2f"%(HS[z] + CS[z] + VM[z] + AMI[z] + ARI[z])) / 5))

                        SC[z] = measures.SC(points[h], x.labelsPred)
                        DI[z] = measures.DI(points[h], x.labelsPred)
                        DB[z] = measures.DB(points[h], x.labelsPred)
                        stdev[z] = measures.stdev(
                            x.bestIndividual, x.labelsPred, k[h], points[h])
                        exSSE[z] = measures.SSE(
                            x.bestIndividual, x.labelsPred, k[h], points[h])
                        exTWCV[z] = measures.TWCV(
                            x.bestIndividual, x.labelsPred, k[h], points[h])

                        executionTime[z] = x.executionTime
                        convergence[z] = x.convergence
                        optimizerName = x.optimizer
                        objfname = x.objfname

                        if(Export_details_labels == True):
                            ExportToFileDetailsLabels = results_directory + "experiment_details_Labels.csv"
                            with open(ExportToFileDetailsLabels, 'a', newline='\n') as out_details_labels:
                                writer_details = csv.writer(
                                    out_details_labels, delimiter=',')
                                # just one time to write the header of the CSV file
                                if (Flag_details_Labels == False):
                                    header_details = numpy.concatenate(
                                        [["Dataset", "Optimizer", "objfname", "k"]])
                                    writer_details.writerow(header_details)
                                    Flag_details_Labels = True
                                a = numpy.concatenate(
                                    [[self.dataset_list[h], optimizerName, objfname, k[h]], x.labelsPred])
                                writer_details.writerow(a)
                            out_details_labels.close()

                        if(Export_details == True):
                            ExportToFileDetails = results_directory + "experiment_details.csv"
                            with open(ExportToFileDetails, 'a', newline='\n') as out_details:
                                writer_details = csv.writer(
                                    out_details, delimiter=',')
                                # just one time to write the header of the CSV file
                                if (Flag_details == False):
                                    if self.labels_exist:
                                        header_details = numpy.concatenate([["Dataset", "Optimizer", "objfname", "k", "ExecutionTime", "SSE", "Purity",
                                                                        "Entropy", "HS", "CS", "VM", "AMI", "ARI", "Fmeasure", "TWCV", "SC", "Accuracy", "DI", "DB", "STDev"], CnvgHeader])
                                    else:
                                        header_details = numpy.concatenate(
                                            [["Dataset", "Optimizer", "objfname", "k", "ExecutionTime", "SSE", "TWCV", "SC", "DI", "DB", "STDev"], CnvgHeader])
                                    writer_details.writerow(header_details)
                                    Flag_details = True
                                if self.labels_exist:
                                    a = numpy.concatenate([[self.dataset_list[h], optimizerName, objfname, k[h], float("%0.2f" % (executionTime[z])), float("%0.2f" % (exSSE[z])), float("%0.2f" % (purity[z])), float("%0.2f" % (entropy[z])), float("%0.2f" % (HS[z])), float("%0.2f" % (CS[z])),  float("%0.2f" % (VM[z])),  float(
                                        "%0.2f" % (AMI[z])),  float("%0.2f" % (ARI[z])), float("%0.2f" % (Fmeasure[z])),  float("%0.2f" % (exTWCV[z])),  float("%0.2f" % (SC[z])),  float("%0.2f" % (accuracy[z])),  float("%0.2f" % (DI[z])), float("%0.2f" % (DB[z])), float("%0.2f" % (stdev[z]))], numpy.around(convergence[z], decimals=2)])
                                else:
                                    a = numpy.concatenate([[self.dataset_list[h], optimizerName, objfname, k[h], float("%0.2f" % (executionTime[z])), float("%0.2f" % (exSSE[z])), float("%0.2f" % (
                                        exTWCV[z])),  float("%0.2f" % (SC[z])),  float("%0.2f" % (DI[z])), float("%0.2f" % (DB[z])), float("%0.2f" % (stdev[z]))], numpy.around(convergence[z], decimals=2)])

                                writer_details.writerow(a)
                            out_details.close()

                    if(Export == True):
                        ExportToFile = results_directory + "experiment.csv"

                        with open(ExportToFile, 'a', newline='\n') as out:
                            writer = csv.writer(out, delimiter=',')
                            if (Flag == False):  # just one time to write the header of the CSV file
                                if self.labels_exist:
                                    header = numpy.concatenate([["Dataset", "Optimizer", "objfname", "k", "ExecutionTime", "SSE", "Purity", "Entropy",
                                                                "HS", "CS", "VM", "AMI", "ARI", "Fmeasure", "TWCV", "SC", "Accuracy", "DI", "DB", "STDev"], CnvgHeader])
                                else:
                                    header = numpy.concatenate([["Dataset", "Optimizer","objfname","k","ExecutionTime","SSE","TWCV","SC","DI","DB","STDev"],CnvgHeader])
                                writer.writerow(header)
                                Flag = True # at least one experiment

                            avgSSE = str(float("%0.2f" %(sum(exSSE) / self.num_of_runs)))
                            avgTWCV = str(float("%0.2f" %(sum(exTWCV) / self.num_of_runs)))
                            avgPurity = str(float("%0.2f" %(sum(purity) / self.num_of_runs)))
                            avgEntropy = str(float("%0.2f" %(sum(entropy) / self.num_of_runs)))
                            avgHomo = str(float("%0.2f" %(sum(HS) / self.num_of_runs)))
                            avgComp = str(float("%0.2f" %(sum(CS) / self.num_of_runs)))
                            avgVmeas = str(float("%0.2f" %(sum(VM) / self.num_of_runs)))
                            avgAMI = str(float("%0.2f" %(sum(AMI) / self.num_of_runs)))
                            avgARI = str(float("%0.2f" %(sum(ARI) / self.num_of_runs)))
                            avgFmeasure = str(float("%0.2f" %(sum(Fmeasure) / self.num_of_runs)))
                            avgSC = str(float("%0.2f" %(sum(SC) / self.num_of_runs)))
                            avgAccuracy = str(float("%0.2f" %(sum(accuracy) / self.num_of_runs)))
                            avgDI = str(float("%0.2f" %(sum(DI) / self.num_of_runs)))
                            avgDB = str(float("%0.2f"%(sum(DB) / self.num_of_runs)) )
                            avgStdev = str(float("%0.2f"%(sum(stdev) / self.num_of_runs)))
                            # avgAgg = str(float("%0.2f"%(sum(Agg) / NumOfRuns)))

                            avgExecutionTime = float("%0.2f" %(sum(executionTime) / self.num_of_runs))
                            avgConvergence = numpy.around(numpy.mean(
                                convergence, axis=0, dtype=numpy.float64), decimals=2).tolist()
                            if self.labels_exist:
                                a = numpy.concatenate([[self.dataset_list[h], optimizerName,objfname,k[h], avgExecutionTime,avgSSE,avgPurity,avgEntropy,avgHomo, avgComp, avgVmeas, avgAMI, avgARI, avgFmeasure, avgTWCV, avgSC, avgAccuracy, avgDI, avgDB, avgStdev],avgConvergence])
                            else:
                                a = numpy.concatenate([[self.dataset_list[h], optimizerName,objfname,k[h], avgExecutionTime,avgSSE,avgTWCV, avgSC, avgDI, avgDB, avgStdev],avgConvergence])
                            writer.writerow(a)
                        out.close()

        if Export_convergence == True:
            conv_plot.run(results_directory, self.optimizer,
                        self.objective_func, self.dataset_list, Iterations)

        if Export_boxplot == True:
            if self.labels_exist:
                ev_measures = ['SSE','Purity','Entropy', 'HS', 'CS', 'VM', 'AMI', 'ARI', 'Fmeasure', 'TWCV', 'SC', 'Accuracy', 'DI', 'DB', 'STDev']
            else:
                ev_measures = ['SSE','TWCV', 'SC', 'DI', 'DB', 'STDev']
            box_plot.run(results_directory, self.optimizer, self.objective_func,
                        self.dataset_list, ev_measures, Iterations)

        print("Execution completed")


    def selector(self, algo, objective_name, k, f, popSize,Iter, points, metric):
        """
        This is used to call the algorithm which is selected

        Parameters
        ----------
        algo : int
            The index of the selected algorithm
        objective_name : str
            The name of the selected function
        k : int
            Number of clusters
        f : int
            Number of features
        popSize : int
            Size of population (the number of individuals at each iteration)
        Iter : int
            The number of iterations
        points : numpy.ndaarray
            The attribute values of all the points

        Returns
        -----------
        obj
            x: solution object returned by the selected algorithm
        """
        lb = 0
        ub = 1

        if(algo =="SSA"):
            x = cssa.SSA(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="PSO"):
            x = cpso.PSO(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="GA"):
            x = cga.GA(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="BAT"):
            x = cbat.BAT(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="FFA"):
            x = cffa.FFA(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="GWO"):
            x = cgwo.GWO(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="WOA"):
            x = cwoa.WOA(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="MVO"):
            x = cmvo.MVO(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="MFO"):
            x = cmfo.MFO(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        if(algo =="CS"):
            x = ccs.CS(getattr(objectives, objective_name),lb,ub,k * f,popSize,Iter, k, points, metric)        
        return x



# -*- coding: utf-8 -*-
"""
Created on Sun May 29 00:49:35 2016

@author: hossam
"""

#% ======================================================== % 
#% Files of the Matlab programs included in the book:       %
#% Xin-She Yang, Nature-Inspired Metaheuristic Algorithms,  %
#% Second Edition, Luniver Press, (2010).   www.luniver.com %
#% ======================================================== %    
#
#% -------------------------------------------------------- %
#% Firefly Algorithm for constrained optimization using     %
#% for the design of a spring (benchmark)                   % 
#% by Xin-She Yang (Cambridge University) Copyright @2009   %
#% -------------------------------------------------------- %

import numpy
import math
import time
from .._solution import solution





def alpha_new(alpha,NGen):
    #% alpha_n=alpha_0(1-delta)^NGen=10^(-4);
    #% alpha_0=0.9
    delta=1-(10**(-4)/0.9)**(1/NGen);
    alpha=(1-delta)*alpha
    return alpha



def FFA(objf,lb,ub,dim,n,MaxGeneration, k, points, metric):

    #General parameters

    #n=50 #number of fireflies
    #dim=30 #dim  
    #lb=-50
    #ub=50
    #MaxGeneration=500
 
    #FFA parameters
    alpha=0.5  # Randomness 0--1 (highly random)
    betamin=0.20  # minimum value of beta
    gamma=1   # Absorption coefficient
    
    
    zn=numpy.ones(n)
    zn.fill(float("inf")) 
    
    
    #ns(i,:)=Lb+(Ub-Lb).*rand(1,d);
    ns = numpy.zeros((n, dim))
    ns=numpy.random.uniform(0,1,(n,dim)) *(ub-lb)+lb
    
    Lightn=numpy.ones(n)
    Lightn.fill(float("inf")) 
    labelsPred=numpy.zeros((n,len(points)))
    
    #[ns,Lightn]=init_ffa(n,d,Lb,Ub,u0)
    
    convergence=[]
    s=solution()

     
    print("CS is optimizing  \""+objf.__name__+"\"")    
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    # Main loop
    for Iteration in range (0,MaxGeneration):     # start iterations
    
        #% This line of reducing alpha is optional
        alpha=alpha_new(alpha,MaxGeneration);
        
        #% Evaluate new solutions (for all n fireflies)
        for i in range(0,n):            
            startpts = numpy.reshape(ns[i,:], (k,(int)(dim/k)))
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
            else:
                fitnessValue, labelsPredValues=objf(startpts, points, k) 
            zn[i] = fitnessValue
            Lightn[i]=zn[i]
            labelsPred[i,:] = labelsPredValues
        
        
                
        
        # Ranking fireflies by their light intensity/objectives
    
        
        Lightn=numpy.sort(zn)
        Index=numpy.argsort(zn)
        ns=ns[Index,:]
        
        
        #Find the current best
        nso=ns
        Lighto=Lightn
        nbest=ns[0,:] 
        Lightbest=Lightn[0]
        labelsPredBest=labelsPred[0]
        
        #% For output only
        fbest=Lightbest
        
        #% Move all fireflies to the better locations
    #    [ns]=ffa_move(n,d,ns,Lightn,nso,Lighto,nbest,...
    #          Lightbest,alpha,betamin,gamma,Lb,Ub);
        
        scale=numpy.ones(dim)*abs(ub-lb)

        for i in range (0,n):
            # The attractiveness parameter beta=exp(-gamma*r)
            for j in range(0,n):
                r=numpy.sqrt(numpy.sum((ns[i,:]-ns[j,:])**2));
                #r=1
                # Update moves
                if Lightn[i]>Lighto[j]: # Brighter and more attractive
                   beta0=1
                   beta=(beta0-betamin)*math.exp(-gamma*r**2)+betamin
                   tmpf=alpha*(numpy.random.rand(dim)-0.5)*scale
                   ns[i,:]=ns[i,:]*(1-beta)+nso[j,:]*beta+tmpf
        
        
        #ns=numpy.clip(ns, lb, ub)
        
        convergence.append(fbest)
        	
        if (Iteration%1==0):
               print(['At iteration '+ str(Iteration)+ ' the best fitness is '+ str(fbest)])
    #    
       ####################### End main loop
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence
    s.optimizer="FFA"
    s.objfname=objf.__name__
    s.labelsPred = numpy.array(labelsPredBest, dtype=numpy.int64)
    s.bestIndividual = nbest
    
    return s
    
    
    
    
    



# -*- coding: utf-8 -*-
"""
Created on Wed May 11 17:06:34 2016

@author: hossam
"""
import random
import numpy
import time
import math
import sklearn
from numpy import asarray
from sklearn.preprocessing import normalize
from .._solution import solution


  



def normr(Mat):
   """normalize the columns of the matrix
   B= normr(A) normalizes the row
   the dtype of A is float"""
   Mat=Mat.reshape(1, -1)
     # Enforce dtype float
   if Mat.dtype!='float':
      Mat = asarray(Mat,dtype=float)

   # if statement to enforce dtype float
   #B = normalize(Mat,norm='l2',axis=1)
   B=(Mat-min(Mat))/(max(Mat)-min(Mat))
   B=numpy.reshape(B,-1)
   return B

def randk(t):
    if (t%2)==0:
        s=0.25
    else:
         s=0.75
    return s

def RouletteWheelSelection(weights):
  accumulation = numpy.cumsum(weights)
  p = random.random() * accumulation[-1]
  chosen_index = -1;
  for index in range (0, len(accumulation)):
    if (accumulation[index] > p):
      chosen_index = index;
      break;
  
  choice = chosen_index;

  return choice



def MVO(objf,lb,ub,dim,N,Max_time, k, points, metric):

    "parameters"
    #dim=30
    #lb=-100
    #ub=100
    WEP_Max=1;
    WEP_Min=0.2
    #Max_time=1000
    #N=50
    
    Universes=numpy.random.uniform(0,1,(N,dim)) *(ub-lb)+lb

    Sorted_universes=numpy.copy(Universes)
    
    labelsPred=numpy.zeros((N,len(points)))
    Sorted_labels=numpy.copy(labelsPred)
    
    convergence=numpy.zeros(Max_time)
     
    
    Best_universe=[0]*dim;
    Best_universe_Inflation_rate= float("inf")
    labelsPredBest=numpy.zeros(len(points))
    
    
    
    
    s=solution()

    
    Time=1;
    ############################################
    print("MVO is optimizing  \""+objf.__name__+"\"")    
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    while (Time<Max_time+1):
    
        "Eq. (3.3) in the paper"
        WEP=WEP_Min+Time*((WEP_Max-WEP_Min)/Max_time)
       
        TDR=1-(math.pow(Time,1/6)/math.pow(Max_time,1/6))
      
        Inflation_rates=[0]*len(Universes)
        
       
       
        for i in range(0,N):
            Universes[i,:]=numpy.clip(Universes[i,:], lb, ub)
    
            
    
            startpts = numpy.reshape(Universes[i,:], (k,(int)(dim/k)))
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
            else:
                fitnessValue, labelsPredValues=objf(startpts, points, k) 
            Inflation_rates[i] = fitnessValue
            labelsPred[i,:] = labelsPredValues
           
       
               
            if Inflation_rates[i]<Best_universe_Inflation_rate :
                        
                Best_universe_Inflation_rate=Inflation_rates[i]
                Best_universe=numpy.array(Universes[i,:])
                labelsPredBest=numpy.array(labelsPred[i,:])
             
        
        sorted_Inflation_rates = numpy.sort(Inflation_rates)
        sorted_indexes = numpy.argsort(Inflation_rates)
        
        for newindex in range(0,N):
            Sorted_universes[newindex,:]=numpy.array(Universes[sorted_indexes[newindex],:])   
            Sorted_labels[newindex,:]=numpy.array(labelsPred[sorted_indexes[newindex],:])   
            
        normalized_sorted_Inflation_rates=numpy.copy(normr(sorted_Inflation_rates))
    
        
        Universes[0,:]= numpy.array(Sorted_universes[0,:])
        labelsPred[0,:]=Sorted_labels[0,:]
    
        for i in range(1,N):
            Back_hole_index=i
            for j in range(0,dim):
                r1=random.random()
                
                if r1<normalized_sorted_Inflation_rates[i]:
                    White_hole_index=RouletteWheelSelection(-sorted_Inflation_rates);
    
                    if White_hole_index==-1:
                        White_hole_index=0;
                    White_hole_index=0;
                    Universes[Back_hole_index,j]=Sorted_universes[White_hole_index,j]
                    labelsPred[Back_hole_index,j]=Sorted_labels[White_hole_index,j]
            
                r2=random.random() 
                
                
                if r2<WEP:
                    r3=random.random() 
                    if r3<0.5:                    
                        Universes[i,j]=Best_universe[j]+TDR*((ub-lb)*random.random()+lb) #random.uniform(0,1)+lb);                    
                    if r3>0.5:          
                        Universes[i,j]=Best_universe[j]-TDR*((ub-lb)*random.random()+lb) #random.uniform(0,1)+lb);
                              
        
        
        
        
        convergence[Time-1]=Best_universe_Inflation_rate
        if (Time%1==0):
               print(['At iteration '+ str(Time)+ ' the best fitness is '+ str(Best_universe_Inflation_rate)]);

        
        
        
        Time=Time+1
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence
    s.optimizer="MVO"
    s.objfname=objf.__name__
    s.bestIndividual = Best_universe
    s.labelsPred = numpy.array(labelsPredBest, dtype=numpy.int64)

    return s






import matplotlib.pyplot as plt
import pandas as pd 
from . import _objectives

def run(results_directory, optimizer, objectivefunc, dataset_List, Iterations):
    plt.ioff()
    fileResultsData = pd.read_csv(results_directory + '/experiment.csv')

    for d in range(len(dataset_List)):
        dataset_filename = dataset_List[d] + '.csv' 
        for j in range (0, len(objectivefunc)):
            objective_name = objectivefunc[j]

            startIteration = 0                
            if 'SSA' in optimizer:
                startIteration = 1             
            allGenerations = [x+1 for x in range(startIteration,Iterations)]   
            for i in range(len(optimizer)):
                optimizer_name = optimizer[i]
                fileResultsData = fileResultsData.drop(['SSE','Purity','Entropy','HS','CS','VM','AMI','ARI','Fmeasure','TWCV','SC','Accuracy','DI','DB','STDev'], errors='ignore', axis=1)
                row = fileResultsData[(fileResultsData["Dataset"] == dataset_List[d]) & (fileResultsData["Optimizer"] == optimizer_name) & (fileResultsData["objfname"] == objective_name)]
                row = row.iloc[:, 5+startIteration:]

                plt.plot(allGenerations, row.values.tolist()[0], label=optimizer_name)
            plt.xlabel('Iterations')
            plt.ylabel('Fitness')
            plt.legend(loc="top right", bbox_to_anchor=(1.2,1.02))
            plt.grid()
            fig_name = results_directory + "/convergence-" + dataset_List[d] + "-" + objective_name + ".png"
            plt.savefig(fig_name, bbox_inches='tight')
            plt.clf()
            #plt.show()


# -*- coding: utf-8 -*-
"""
Created on Mon May 16 00:27:50 2016

@author: Hossam Faris
"""

import random
import numpy
import math
from .._solution import solution
import time


    

def GWO(objf,lb,ub,dim,SearchAgents_no,Max_iter, k, points, metric):
    
    
    #Max_iter=1000
    #lb=-100
    #ub=100
    #dim=30  
    #SearchAgents_no=5
    
    # initialize alpha, beta, and delta_pos
    Alpha_score=float("inf")
    Alpha_pos=numpy.zeros(dim)
    Alpha_labels=numpy.zeros(dim)
    
    Beta_score=float("inf")
    Beta_pos=numpy.zeros(dim)
    Beta_labels=numpy.zeros(dim)
    
    Delta_score=float("inf")
    Delta_pos=numpy.zeros(dim)
    Delta_labels=numpy.zeros(dim)

    #Initialize the positions of search agents
    Positions = numpy.zeros((SearchAgents_no, dim))
    Positions=numpy.random.uniform(0,1,(SearchAgents_no,dim)) *(ub-lb)+lb
    labelsPred=numpy.zeros((SearchAgents_no,len(points)))
    
    Convergence_curve=numpy.zeros(Max_iter)
    s=solution()

     # Loop counter
    print("GWO is optimizing  \""+objf.__name__+"\"")    
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    # Main loop
    for l in range(0,Max_iter):
        for i in range(0,SearchAgents_no):
            
            # Return back the search agents that go beyond the boundaries of the search space
            Positions[i,:]=numpy.clip(Positions[i,:], lb, ub)

            # Calculate objective function for each search agent
            startpts = numpy.reshape(Positions[i,:], (k,(int)(dim/k)))
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
            else:
                fitnessValue, labelsPredValues=objf(startpts, points, k) 
            fitness = fitnessValue
            labelsPred[i,:] = labelsPredValues
            
            # Update Alpha, Beta, and Delta
            if fitness<Alpha_score :
                Delta_score=Beta_score  # Update delte
                Delta_pos=Beta_pos.copy()
                Delta_labels=Beta_labels.copy()
                Beta_score=Alpha_score  # Update beta
                Beta_pos=Alpha_pos.copy()
                Beta_labels=Alpha_labels.copy()
                Alpha_score=fitness; # Update alpha
                Alpha_pos=Positions[i,:].copy()
                Alpha_labels=labelsPred[i,:].copy()
                        
            if (fitness>Alpha_score and fitness<Beta_score ):
                Delta_score=Beta_score  # Update delte
                Delta_pos=Beta_pos.copy()
                Delta_labels=Beta_labels.copy()
                Beta_score=fitness  # Update beta
                Beta_pos=Positions[i,:].copy()
                Beta_labels=labelsPred[i,:].copy()            
            
            if (fitness>Alpha_score and fitness>Beta_score and fitness<Delta_score):                 
                Delta_score=fitness # Update delta
                Delta_pos=Positions[i,:].copy()
                Delta_labels=labelsPred[i,:].copy()           
                
        
        a=2-l*((2)/Max_iter); # a decreases linearly fron 2 to 0
        
        # Update the Position of search agents including omegas
        for i in range(0,SearchAgents_no):
            for j in range (0,dim):     
                           
                r1=random.random() # r1 is a random number in [0,1]
                r2=random.random() # r2 is a random number in [0,1]
                
                A1=2*a*r1-a; # Equation (3.3)
                C1=2*r2; # Equation (3.4)
                
                D_alpha=abs(C1*Alpha_pos[j]-Positions[i,j]); # Equation (3.5)-part 1
                X1=Alpha_pos[j]-A1*D_alpha; # Equation (3.6)-part 1
                           
                r1=random.random()
                r2=random.random()
                
                A2=2*a*r1-a; # Equation (3.3)
                C2=2*r2; # Equation (3.4)
                
                D_beta=abs(C2*Beta_pos[j]-Positions[i,j]); # Equation (3.5)-part 2
                X2=Beta_pos[j]-A2*D_beta; # Equation (3.6)-part 2       
                
                r1=random.random()
                r2=random.random() 
                
                A3=2*a*r1-a; # Equation (3.3)
                C3=2*r2; # Equation (3.4)
                
                D_delta=abs(C3*Delta_pos[j]-Positions[i,j]); # Equation (3.5)-part 3
                X3=Delta_pos[j]-A3*D_delta; # Equation (3.5)-part 3             
                
                Positions[i,j]=(X1+X2+X3)/3  # Equation (3.7)
                
            
        
        
        Convergence_curve[l]=Alpha_score;

        if (l%1==0):
               print(['At iteration '+ str(l)+ ' the best fitness is '+ str(Alpha_score)]);
    
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=Convergence_curve
    s.optimizer="GWO"
    s.objfname=objf.__name__
    s.labelsPred = numpy.array(Alpha_labels, dtype=numpy.int64)
    s.bestIndividual = Alpha_pos
    
    
    
    
    return s


# -*- coding: utf-8 -*-
"""
Created on Mon May 16 14:19:49 2016

@author: hossam
"""
import random
import numpy
import math
from .._solution import solution
import time




def WOA(objf,lb,ub,dim,SearchAgents_no,Max_iter,k,points, metric):


    #dim=30
    #SearchAgents_no=50
    #lb=-100
    #ub=100
    #Max_iter=500
      
    # initialize position vector and score for the leader
    Leader_pos=numpy.zeros(dim)
    Leader_score=float("inf")  #change this to -inf for maximization problems
    
    
    #Initialize the positions of search agents
    Positions = numpy.zeros((SearchAgents_no, dim))
    
    Positions=numpy.random.uniform(0,1,(SearchAgents_no,dim)) *(ub-lb)+lb
    labelsPred=numpy.zeros((SearchAgents_no,len(points)))
    
    #Initialize convergence
    convergence_curve=numpy.zeros(Max_iter)
    
    
    ############################
    s=solution()

    print("WOA is optimizing  \""+objf.__name__+"\"")    

    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    ############################
    
    t=0  # Loop counter
    
    # Main loop
    while t<Max_iter:
        for i in range(0,SearchAgents_no):
            
            # Return back the search agents that go beyond the boundaries of the search space
            
            #Positions[i,:]=checkBounds(Positions[i,:],lb,ub)            
            Positions[i,:]=numpy.clip(Positions[i,:], lb, ub)

            
            # Calculate objective function for each search agent
            startpts = numpy.reshape(Positions[i,:], (k,(int)(dim/k)))

            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
            else:
                fitnessValue, labelsPredValues=objf(startpts, points, k) 
                
            fitness = fitnessValue
            labelsPred[i,:] = labelsPredValues
            
            # Update the leader
            if fitness<Leader_score: # Change this to > for maximization problem
                Leader_score=fitness; # Update alpha
                Leader_pos=Positions[i,:].copy() # copy current whale position into the leader position
                Leader_labels=labelsPred[i,:].copy() # copy current whale position into the leader position
            
            
        
        
        a=2-t*((2)/Max_iter); # a decreases linearly fron 2 to 0 in Eq. (2.3)
        
        # a2 linearly decreases from -1 to -2 to calculate t in Eq. (3.12)
        a2=-1+t*((-1)/Max_iter);
        
        # Update the Position of search agents 
        for i in range(0,SearchAgents_no):
            r1=random.random() # r1 is a random number in [0,1]
            r2=random.random() # r2 is a random number in [0,1]
            
            A=2*a*r1-a  # Eq. (2.3) in the paper
            C=2*r2      # Eq. (2.4) in the paper
            
            
            b=1;               #  parameters in Eq. (2.5)
            l=(a2-1)*random.random()+1   #  parameters in Eq. (2.5)
            
            p = random.random()        # p in Eq. (2.6)
            
            for j in range(0,dim):
                
                if p<0.5:
                    if abs(A)>=1:
                        rand_leader_index = math.floor(SearchAgents_no*random.random());
                        X_rand = Positions[rand_leader_index, :]
                        D_X_rand=abs(C*X_rand[j]-Positions[i,j]) 
                        Positions[i,j]=X_rand[j]-A*D_X_rand      
                        
                    elif abs(A)<1:
                        D_Leader=abs(C*Leader_pos[j]-Positions[i,j]) 
                        Positions[i,j]=Leader_pos[j]-A*D_Leader      
                    
                    
                elif p>=0.5:
                  
                    distance2Leader=abs(Leader_pos[j]-Positions[i,j])
                    # Eq. (2.5)
                    Positions[i,j]=distance2Leader*math.exp(b*l)*math.cos(l*2*math.pi)+Leader_pos[j]
                    
      
        
        convergence_curve[t]=Leader_score
        if (t%1==0):
               print(['At iteration '+ str(t)+ ' the best fitness is '+ str(Leader_score)]);
        t=t+1
    
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence_curve
    s.optimizer="WOA"   
    s.objfname=objf.__name__
    s.best = Leader_score
    s.bestIndividual = Leader_pos
    s.labelsPred = numpy.array(Leader_labels, dtype=numpy.int64)
    
    
    
    return s





# -*- coding: utf-8 -*-
"""
Created on Sun Mar 31 21:22:53 2019

@author: Raneem
"""


from sklearn import metrics
from sklearn.metrics.pairwise import euclidean_distances
import statistics
import math  
import numpy
import sys
import math

def HS(labelsTrue, labelsPred):    
    return float("%0.2f"%metrics.homogeneity_score(labelsTrue,labelsPred))

def CS(labelsTrue, labelsPred):    
    return float("%0.2f"%metrics.completeness_score(labelsTrue,labelsPred))
    
def VM(labelsTrue, labelsPred):
    return float("%0.2f"%metrics.v_measure_score(labelsTrue,labelsPred))
    
def AMI(labelsTrue, labelsPred):
    return float("%0.2f"%metrics.adjusted_mutual_info_score(labelsTrue,labelsPred))
    
def ARI(labelsTrue, labelsPred):
    return float("%0.2f"%metrics.adjusted_rand_score(labelsTrue,labelsPred))
    
def Fmeasure(labelsTrue, labelsPred):
    return float("%0.2f"%metrics.f1_score(labelsTrue, labelsPred, average='macro'))

def SC(points, labelsPred):#Silhouette Coefficient

    if numpy.unique(labelsPred).size == 1:
        fitness = sys.float_info.max
    else:    
        silhouette=  float("%0.2f"%metrics.silhouette_score(points, labelsPred, metric='euclidean'))
        silhouette = (silhouette + 1) / 2
        fitness = 1 - silhouette
    return fitness

def accuracy(labelsTrue, labelsPred):#Silhouette Coefficient
    #silhouette = metrics.accuracy_score(labelsTrue, labelsPred, normalize=False)
    return ARI(labelsTrue, labelsPred)


def delta_fast(ck, cl, distances):
    values = distances[numpy.where(ck)][:, numpy.where(cl)]
    values = values[numpy.nonzero(values)]

    return numpy.min(values)
    
def big_delta_fast(ci, distances):
    values = distances[numpy.where(ci)][:, numpy.where(ci)]
    #values = values[numpy.nonzero(values)]
            
    return numpy.max(values)

def dunn_fast(points, labels):
    """ Dunn index - FAST (using sklearn pairwise euclidean_distance function)
    
    Parameters
    ----------
    points : numpy.array
        numpy.array([N, p]) of all points
    labels: numpy.array
        numpy.array([N]) labels of all points
    """
    distances = euclidean_distances(points)
    ks = numpy.sort(numpy.unique(labels))
    
    deltas = numpy.ones([len(ks), len(ks)])*1000000
    big_deltas = numpy.zeros([len(ks), 1])
    
    l_range = list(range(0, len(ks)))
    
    for k in l_range:
        for l in (l_range[0:k]+l_range[k+1:]):
            deltas[k, l] = delta_fast((labels == ks[k]), (labels == ks[l]), distances)
        
        big_deltas[k] = big_delta_fast((labels == ks[k]), distances)

    di = numpy.min(deltas)/numpy.max(big_deltas)
    return di
    

def DI(points, labelsPred):#dunn index
    dunn = float("%0.2f"%dunn_fast(points, labelsPred))
    if(dunn < 0):
        dunn = 0
    fitness = 1 - dunn
    return fitness

    
def DB(points, labelsPred):
    return float("%0.2f"%metrics.davies_bouldin_score(points, labelsPred))

def stdev(individual, labelsPred, k, points):
    std = 0    
    distances = []
    f = (int)(len(individual) / k)
    startpts = numpy.reshape(individual, (k,f))   

    for i in range(k):
        index_list = numpy.where(labelsPred == i)
        distances = numpy.append(distances, numpy.linalg.norm(points[index_list]-startpts[i], axis = 1))
    
    std =  numpy.std(distances)
        
    #stdev = math.sqrt(std)/ k 
    #print("stdev:",stdev)
    return std
    
    
'''
def SSE(individual, k, points):
    
    f = (int)(len(individual) / k)
    startpts = numpy.reshape(individual, (k,f))    
    labelsPred = [-1] * len(points)
    sse = 0
    
    for i in range(len(points)):
        distances = numpy.linalg.norm(points[i]-startpts, axis = 1)
        sse = sse + numpy.min(distances)
        clust = numpy.argmin(distances)
        labelsPred[i] = clust
        
    if numpy.unique(labelsPred).size < k:
        sse = sys.float_info.max
               
    print("SSE:",sse)
    return sse
'''

def SSE(individual, labelsPred, k, points):
    
    f = (int)(len(individual) / k)
    startpts = numpy.reshape(individual, (k,f)) 
    fitness = 0
       
    
    centroidsForPoints = startpts[labelsPred]
    fitnessValues = numpy.linalg.norm(points-centroidsForPoints, axis = 1)**2
    fitness = sum(fitnessValues)
    return fitness


def TWCV(individual, labelsPred, k, points):    
    sumAllFeatures = sum(sum(numpy.power(points,2)))
    sumAllPairPointsCluster = 0
    for clusterId in range(k):
        indices = numpy.where(numpy.array(labelsPred) == clusterId)[0]
        pointsInCluster = points[numpy.array(indices)]
        sumPairPointsCluster = sum(pointsInCluster)
        sumPairPointsCluster = numpy.power(sumPairPointsCluster,2)
        if len(pointsInCluster) != 0:
            sumPairPointsCluster = sum(sumPairPointsCluster)
            sumPairPointsCluster = sumPairPointsCluster/len(pointsInCluster)
        
        sumAllPairPointsCluster += sumPairPointsCluster
    fitness = (sumAllFeatures - sumAllPairPointsCluster)
    return fitness


def purity(labelsTrue,labelsPred):            
    # get the set of unique cluster ids
    labelsTrue=numpy.asarray(labelsTrue).astype(int)
    labelsPred=numpy.asarray(labelsPred).astype(int)
    
    k=(max(labelsTrue)+1).astype(int)
    
    totalSum = 0;
    
    for i in range(0,k):
        max_freq=0
    
        t1=numpy.where(labelsPred == i)

        for j in range(0,k):
            t2=numpy.where(labelsTrue == j)
            z=numpy.intersect1d(t1,t2);
           
            e=numpy.shape(z)[0]
               
            if (e >= max_freq):
                max_freq=e
             
        totalSum=totalSum + max_freq

    purity=totalSum/numpy.shape(labelsTrue)[0]
     
    #print("purity:",purity)
    
    return purity

def entropy(labelsTrue,labelsPred):            
        # get the set of unique cluster ids
    labelsTrue=numpy.asarray(labelsTrue).astype(int)
    labelsPred=numpy.asarray(labelsPred).astype(int)
    
    k=(max(labelsTrue)+1).astype(int)
    
    entropy=0
       
    for i in range(0,k):
        
        t1=numpy.where(labelsPred == i)

        entropyI=0

        for j in range(0,k):
            t2=numpy.where(labelsTrue == j)
            z=numpy.intersect1d(t1,t2);
           
            e=numpy.shape(z)[0]
            if (e!=0):
                entropyI=entropyI+(e/numpy.shape(t1)[1])*math.log(e/numpy.shape(t1)[1])
               
        a=numpy.shape(t1)[1]
        b=numpy.shape(labelsTrue)[0]
        entropy=entropy+(( a / b )*((-1 / math.log(k))*entropyI)) 
    
    #print("entropy:",entropy)
     
    return entropy


from .CSSA import SSA 
from .CPSO import PSO 
from .CGA import GA 
from .CBAT import BAT 
from .CFFA import FFA 
from .CGWO import GWO 
from .CWOA import WOA 
from .CMVO import MVO 
from .CMFO import MFO 
from .CCS import CS


# -*- coding: utf-8 -*-
"""
Created on Mon May 16 10:42:18 2016

@author: hossam
"""

import random
import numpy
import math
from .._solution import solution
import time



  
def MFO(objf,lb,ub,dim,N,Max_iteration,k,points, metric):


    #Max_iteration=1000
    #lb=-100
    #ub=100
    #dim=30
    N=50 # Number of search agents
    
    
    
    #Initialize the positions of moths
    Moth_pos=numpy.random.uniform(0,1,(N,dim)) *(ub-lb)+lb
    Moth_fitness=numpy.full(N,float("inf"))
    Moth_labels=numpy.zeros((N,len(points)))
    #Moth_fitness=numpy.fell(float("inf"))
    
    Convergence_curve=numpy.zeros(Max_iteration)
    
    
    sorted_population=numpy.copy(Moth_pos)
    sorted_labels=numpy.copy(Moth_labels)
    fitness_sorted=numpy.zeros(N)
    #####################
    best_flames=numpy.copy(Moth_pos)
    best_labels=numpy.copy(Moth_labels)
    best_flame_fitness=numpy.zeros(N)
    ####################
    double_population=numpy.zeros((2*N,dim))
    double_labels=numpy.zeros((2*N,len(points)))
    double_fitness=numpy.zeros(2*N)
    
    double_sorted_population=numpy.zeros((2*N,dim))
    double_sorted_labels=numpy.zeros((2*N,len(points)))
    double_fitness_sorted=numpy.zeros(2*N)
    #########################
    previous_population=numpy.zeros((N,dim))
    previous_labels=numpy.zeros((N,len(points)))
    previous_fitness=numpy.zeros(N)


    s=solution()

    print("MFO is optimizing  \""+objf.__name__+"\"")    

    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    Iteration=1;    
    
    # Main loop
    while (Iteration<Max_iteration):
        
        # Number of flames Eq. (3.14) in the paper
        Flame_no=round(N-Iteration*((N-1)/Max_iteration));
        
        for i in range(0,N):
            
            # Check if moths go out of the search spaceand bring it back
            Moth_pos[i,:]=numpy.clip(Moth_pos[i,:], lb, ub)

            # evaluate moths
            startpts = numpy.reshape(Moth_pos[i,:], (k,(int)(dim/k)))
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
            else:
                fitnessValue, labelsPredValues=objf(startpts, points, k) 
            Moth_fitness[i] = fitnessValue
            Moth_labels[i,:] = labelsPredValues
           
            
        
           
        if Iteration==1:
            # Sort the first population of moths
            fitness_sorted=numpy.sort(Moth_fitness)
            I=numpy.argsort(Moth_fitness)
            
            sorted_population=Moth_pos[I,:]
            sorted_labels=Moth_labels[I,:]
               
            
            #Update the flames
            best_flames=sorted_population
            best_flame_fitness=fitness_sorted
            best_labels=sorted_labels
        else:
    #        
    #        # Sort the moths
            double_population=numpy.concatenate((previous_population,best_flames),axis=0)
            double_labels=numpy.concatenate((previous_labels,best_labels),axis=0)
            double_fitness=numpy.concatenate((previous_fitness, best_flame_fitness),axis=0);
    #        
            double_fitness_sorted =numpy.sort(double_fitness);
            I2 =numpy.argsort(double_fitness);
    #        
    #        
            for newindex in range(0,2*N):
                double_sorted_population[newindex,:]=numpy.array(double_population[I2[newindex],:])           
                double_sorted_labels[newindex,:]=numpy.array(double_labels[I2[newindex],:])           
            
            fitness_sorted=double_fitness_sorted[0:N]
            sorted_population=double_sorted_population[0:N,:]
            sorted_labels=double_sorted_labels[0:N,:]
    #        
    #        # Update the flames
            best_flames=sorted_population;
            best_labels=sorted_labels;
            best_flame_fitness=fitness_sorted;
    
    #    
    #   # Update the position best flame obtained so far
        Best_flame_score=fitness_sorted[0]
        Best_flame_pos=sorted_population[0,:]
        Best_labelsPred=sorted_labels[0,:]
    #      
        previous_population=Moth_pos;
        previous_labels=Moth_labels;
        previous_fitness=Moth_fitness;
    #    
        # a linearly dicreases from -1 to -2 to calculate t in Eq. (3.12)
        a=-1+Iteration*((-1)/Max_iteration);
        

        
        # Loop counter
        for i in range(0,N):
    #        
            for j in range(0,dim):
                if (i<=Flame_no): #Update the position of the moth with respect to its corresponsing flame
    #                
                    # D in Eq. (3.13)
                    distance_to_flame=abs(sorted_population[i,j]-Moth_pos[i,j])
                    b=1
                    t=(a-1)*random.random()+1;
    #                
    #                % Eq. (3.12)
                    Moth_pos[i,j]=distance_to_flame*math.exp(b*t)*math.cos(t*2*math.pi)+sorted_population[i,j]
    #            end
    #            
                if i>Flame_no: # Upaate the position of the moth with respct to one flame
    #                
    #                % Eq. (3.13)
                    distance_to_flame=abs(sorted_population[i,j]-Moth_pos[i,j]);
                    b=1;
                    t=(a-1)*random.random()+1;
    #                
    #                % Eq. (3.12)
                    Moth_pos[i,j]=distance_to_flame*math.exp(b*t)*math.cos(t*2*math.pi)+sorted_population[Flame_no,j]
        
        Convergence_curve[Iteration]=Best_flame_score
        #Display best fitness along the iteration
        if (Iteration%1==0):
            print(['At iteration '+ str(Iteration)+ ' the best fitness is '+ str(Best_flame_score)]);
    
    
    
    
        Iteration=Iteration+1; 
    
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=Convergence_curve
    s.optimizer="MFO"   
    s.objfname=objf.__name__
    s.bestIndividual = Best_flame_pos
    s.labelsPred = numpy.array(Best_labelsPred, dtype=numpy.int64)
    
    
    
    return s
    







# -*- coding: utf-8 -*-
"""
Created on Tue May 17 17:06:19 2016

@author: Hossam Faris
"""

class solution:
    def __init__(self):
        self.best = 0
        self.bestIndividual=[]
        self.convergence = []
        self.optimizer=""
        self.objfname=""
        self.startTime=0
        self.endTime=0
        self.executionTime=0
        self.lb=0
        self.ub=0
        self.dim=0
        self.popnum=0
        self.maxiers=0
        self.k = 2
        self.points = []
        self.labelsPred = []


"""
Created on Sat Feb  24 20:18:05 2019

@author: Raneem
"""
import numpy
import random
import time
import sys

from .._solution import solution

def crossoverPopulaton(population, scores, popSize, crossoverProbability, keep):
    """    
    The crossover of all individuals
    
    Parameters
    ---------- 
    population : list
        The list of individuals
    scores : list
        The list of fitness values for each individual
    popSize: int
        Number of chrmosome in a population  
    crossoverProbability: float
        The probability of crossing a pair of individuals
    keep: int
        Number of best individuals to keep without mutating for the next generation
         
    Returns
    -------
    N/A
    """
    #initialize a new population
    
    newPopulation = numpy.empty_like(population)
    newPopulation[0:keep] = population[0:keep]
    #Create pairs of parents. The number of pairs equals the number of individuals divided by 2
    for i  in range(keep, popSize, 2):
        #pair of parents selection        
        
        parent1, parent2 = pairSelection(population, scores, popSize)
        crossoverLength = min(len(parent1), len(parent2))
        parentsCrossoverProbability = random.uniform(0.0, 1.0)
        if parentsCrossoverProbability < crossoverProbability:
            offspring1, offspring2 = crossover(crossoverLength, parent1, parent2)
        else:
            offspring1 = parent1.copy()
            offspring2 = parent2.copy()
    
        #Add offsprings to population
        newPopulation[i] = numpy.copy(offspring1)
        newPopulation[i + 1] = numpy.copy(offspring2)
     
    return newPopulation
        
    
def mutatePopulaton(population, popSize, mutationProbability, keep, lb, ub):
    """    
    The mutation of all individuals
    
    Parameters
    ---------- 
    population : list
        The list of individuals
    popSize: int
        Number of chrmosome in a population  
    mutationProbability: float
        The probability of mutating an individual
    keep: int
        Number of best individuals to keep without mutating for the next generation
    lb: list
        lower bound limit list
    ub: list
        Upper bound limit list
         
    Returns
    -------
    N/A
    """
    for i  in range(keep, popSize):
        #Mutation   
        offspringMutationProbability = random.uniform(0.0, 1.0)
        if offspringMutationProbability < mutationProbability:
            mutation(population[i], len(population[i]), lb, ub)
        

def elitism(population, scores, bestIndividual, bestScore):
    """    
    This melitism operator of the population
    
    Parameters
    ----------    
    population : list
        The list of individuals
    scores : list
        The list of fitness values for each individual
    bestIndividual : list
        An individual of the previous generation having the best fitness value          
    bestScore : float
        The best fitness value of the previous generation        
    
    Returns
    -------
    N/A
    """
    
    # get the worst individual
    worstFitnessId = selectWorstIndividual(scores)
    
    #replace worst cromosome with best one from previous generation if its fitness is less than the other
    if scores[worstFitnessId] > bestScore:
       population[worstFitnessId] = numpy.copy(bestIndividual)
       scores[worstFitnessId] = numpy.copy(bestScore)
       
    
def selectWorstIndividual(scores):
    """    
    It is used to get the worst individual in a population based n the fitness value
    
    Parameters
    ---------- 
    scores : list
        The list of fitness values for each individual
        
    Returns
    -------
    int
        maxFitnessId: The individual id of the worst fitness value
    """
    
    maxFitnessId = numpy.where(scores == numpy.max(scores))
    maxFitnessId = maxFitnessId[0][0]
    return maxFitnessId

def pairSelection(population, scores, popSize):    
    """    
    This is used to select one pair of parents using roulette Wheel Selection mechanism
    
    Parameters
    ---------- 
    population : list
        The list of individuals
    scores : list
        The list of fitness values for each individual
    popSize: int
        Number of chrmosome in a population
          
    Returns
    -------
    list
        parent1: The first parent individual of the pair
    list
        parent2: The second parent individual of the pair
    """
    parent1Id = rouletteWheelSelectionId(scores, popSize)
    parent1 = population[parent1Id].copy()
    
    parent2Id = rouletteWheelSelectionId(scores, popSize)    
    parent2 = population[parent2Id].copy()
   
    return parent1, parent2
    
def rouletteWheelSelectionId(scores, popSize): 
    """    
    A roulette Wheel Selection mechanism for selecting an individual
    
    Parameters
    ---------- 
    scores : list
        The list of fitness values for each individual
    popSize: int
        Number of chrmosome in a population
          
    Returns
    -------
    id
        individualId: The id of the individual selected
    """
    
    ##reverse score because minimum value should have more chance of selection
    reverse = max(scores) + min(scores)
    if reverse == float("inf"):
        return random.randint(0, popSize - 1)
    reverseScores = reverse - scores.copy()
    sumScores = sum(reverseScores)
    if sumScores == float("inf"):
        return random.randint(0, popSize - 1)
    pick    = random.uniform(0, sumScores)
    current = 0
    for individualId in range(popSize):
        current += reverseScores[individualId]
        if current > pick:
            return individualId

def crossover(individualLength, parent1, parent2):
    """    
    The crossover operator of a two individuals
    
    Parameters
    ---------- 
    individualLength: int
        The maximum index of the crossover
    parent1 : list
        The first parent individual of the pair
    parent2 : list
        The second parent individual of the pair
          
    Returns
    -------
    list
        offspring1: The first updated parent individual of the pair
    list
        offspring2: The second updated parent individual of the pair
    """
    # The point at which crossover takes place between two parents. 
    crossover_point = random.randint(0, individualLength - 1)
    # The new offspring will have its first half of its genes taken from the first parent and second half of its genes taken from the second parent.
    offspring1 = numpy.concatenate([parent1[0:crossover_point],parent2[crossover_point:]])
    # The new offspring will have its first half of its genes taken from the second parent and second half of its genes taken from the first parent.
    offspring2 = numpy.concatenate([parent2[0:crossover_point],parent1[crossover_point:]])
    
    return offspring1, offspring2


def mutation(offspring, individualLength, lb, ub):
    """    
    The mutation operator of a single individual
    
    Parameters
    ---------- 
    offspring : list
        A generated individual after the crossover
    individualLength: int
        The maximum index of the crossover
    lb: list
        lower bound limit list
    ub: list
        Upper bound limit list
         
    Returns
    -------
    N/A
    """
    
    mutationIndex = random.randint(0, individualLength - 1)
    mutationValue = random.uniform(lb, ub)
    offspring[mutationIndex] = mutationValue


def clearDups(Population, lb, ub):    
        
    """    
    It removes individuals duplicates and replace them with random ones
    
    Parameters
    ----------    
    objf : function
        The objective function selected
    lb: list
        lower bound limit list
    ub: list
        Upper bound limit list
    
    Returns
    -------
    list
        newPopulation: the updated list of individuals
    """ 
    newPopulation = numpy.unique(Population, axis=0)
    oldLen = len(Population)
    newLen = len(newPopulation)
    if newLen < oldLen:
        nDuplicates = oldLen - newLen
        newPopulation = numpy.append(newPopulation, numpy.random.uniform(0,1,(nDuplicates,len(Population[0]))) * (numpy.array(ub) - numpy.array(lb)) + numpy.array(lb), axis=0)
            		        
    return newPopulation

def calculateCost(objf, population, dim, popSize, lb, ub, k, points, metric):
        
    """    
    It calculates the fitness value of each individual in the population
    
    Parameters
    ----------    
    objf : function
        The objective function selected    
    population : list
        The list of individuals
    popSize: int
        Number of chrmosomes in a population
    lb: list
        lower bound limit list
    ub: list
        Upper bound limit list
    
    Returns
    -------
    list
        scores: fitness values of all individuals in the population
    """ 
    scores = numpy.full(popSize, float("inf"))
    bestScore = float("inf")
    bestLabelsPred=numpy.full(len(points), numpy.inf)
    
    #Loop through individuals in population
    for i in range(0,popSize):
        # Return back the search agents that go beyond the boundaries of the search space
        population[i] = numpy.clip(population[i], lb, ub)
			
        # Calculate objective function for each search agent
        #scores[i] = objf(population[i,:]) 
        
        startpts = numpy.reshape(population[i,:], (k,(int)(dim/k)))

        if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
            fitness, labelsPred=objf(startpts, points, k, metric) 
        else:
            fitness, labelsPred=objf(startpts, points, k) 
        
        scores[i] = fitness
        if fitness < bestScore:
            bestLabelsPred = labelsPred
            bestScore = fitness
            bestIndividual = population[i]
        
    return scores, bestLabelsPred, bestIndividual
        

def sortPopulation(population, scores):
    """    
    This is used to sort the population according to the fitness values of the individuals
    
    Parameters
    ---------- 
    population : list
        The list of individuals
    scores : list
        The list of fitness values for each individual
          
    Returns
    -------
    list
        population: The new sorted list of individuals
    list
        scores: The new sorted list of fitness values of the individuals
    """
    sortedIndices = scores.argsort()
    population = population[sortedIndices]
    scores = scores[sortedIndices]
    
    return population, scores


def GA(objf,lb,ub,dim,popSize,iters, k, points, metric):
        
    """    
    This is the main method which implements GA
    
    Parameters
    ----------    
    objf : function
        The objective function selected
    lb: list
        lower bound limit list
    ub: list
        Upper bound limit list
    dim: int
        The dimension of the indivisual
    popSize: int
        Number of chrmosomes in a population
    iters: int
        Number of iterations / generations of GA
    
    Returns
    -------
    obj
        s: The solution obtained from running the algorithm
    """
    
    cp = 1 #crossover Probability
    mp = 0.01 #Mutation Probability
    keep = 2; # elitism parameter: how many of the best individuals to keep from one generation to the next
    
    s=solution()
	
        
    bestIndividual=numpy.zeros(dim)    
    bestScore=float("inf")
    bestLabelsPred=numpy.full(len(points), numpy.inf)

    ga = numpy.random.uniform(0,1,(popSize,dim)) *(ub-lb)+lb
    scores=numpy.full(popSize,float("inf"))
    
    for i in range(dim):
        ga[:, i]=numpy.random.uniform(0,1,popSize) * (ub - lb) + lb
    
    convergence_curve=numpy.zeros(iters)
    
    print("GA is optimizing  \""+objf.__name__+"\"")  
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    for l in range(iters):

        #crossover
        ga = crossoverPopulaton(ga, scores, popSize, cp, keep)
           
        #mutation
        mutatePopulaton(ga, popSize, mp, keep, lb, ub)
           
        ga = clearDups(ga, lb, ub)
        
        scores, bestLabelsPred, bestIndividual = calculateCost(objf, ga, dim, popSize, lb, ub, k, points, metric)
            
        bestScore = min(scores)
         
    
        
        #Sort from best to worst
        ga, scores = sortPopulation(ga, scores)
         
        convergence_curve[l]=bestScore     
        
        if (l%1==0):
            print(['At iteration '+ str(l+1)+ ' the best fitness is '+ str(bestScore)]);
        
    bestLabelsPred = numpy.asarray(bestLabelsPred)
    timerEnd=time.time()  
    s.bestIndividual = bestIndividual
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence_curve
    s.optimizer="GA"
    s.labelsPred = numpy.array(bestLabelsPred, dtype=numpy.int64)
    s.objfname=objf.__name__

    return s


# -*- coding: utf-8 -*-
"""
Created on Tue May 24 13:13:28 2016

@author: Hossam Faris
"""
import math
import numpy
import random
import time
from .._solution import solution


    
def get_cuckoos(nest,best,lb,ub,n,dim):
    
    # perform Levy flights
    tempnest=numpy.zeros((n,dim))
    tempnest=numpy.array(nest)
    beta=3/2;
    sigma=(math.gamma(1+beta)*math.sin(math.pi*beta/2)/(math.gamma((1+beta)/2)*beta*2**((beta-1)/2)))**(1/beta);

    s=numpy.zeros(dim)
    for j in range (0,n):
        s=nest[j,:]
        u=numpy.random.randn(len(s))*sigma
        v=numpy.random.randn(len(s))
        step=u/abs(v)**(1/beta)
 
        stepsize=0.01*(step*(s-best))

        s=s+stepsize*numpy.random.randn(len(s))
    
        tempnest[j,:]=numpy.clip(s, lb, ub)

    return tempnest

def get_best_nest(nest,labelsPred, newnest,fitness,n,dim,objf, k, points, metric):
# Evaluating all new solutions
    tempnest=numpy.zeros((n,dim))
    tempnest=numpy.copy(nest)
    templabels=numpy.copy(labelsPred)

    for j in range(0,n):
    #for j=1:size(nest,1),
        startpts = numpy.reshape(newnest[j,:], (k,(int)(dim/k)))
        if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
            fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
        else:
            fitnessValue, labelsPredValues=objf(startpts, points, k) 
        fnew= fitnessValue
        newLabels = labelsPredValues
        
        if fnew<=fitness[j]:
           fitness[j]=fnew
           tempnest[j,:]=newnest[j,:]
           templabels[j,:]=newLabels
        
    # Find the current best

    fmin = min(fitness)
    I=numpy.argmin(fitness)
    bestlocal=tempnest[I,:]
    bestlabels=templabels[I,:]

    return fmin,bestlocal,bestlabels,tempnest,fitness, templabels

# Replace some nests by constructing new solutions/nests
def empty_nests(nest,pa,n,dim):

    # Discovered or not 
    tempnest=numpy.zeros((n,dim))

    K=numpy.random.uniform(0,1,(n,dim))>pa
    
    
    stepsize=random.random()*(nest[numpy.random.permutation(n),:]-nest[numpy.random.permutation(n),:])

    
    tempnest=nest+stepsize*K
 
    return tempnest
##########################################################################


def CS(objf,lb,ub,dim,n,N_IterTotal,k,points, metric):

    #lb=-1
    #ub=1
    #n=50
    #N_IterTotal=1000
    #dim=30
    
    # Discovery rate of alien eggs/solutions
    pa=0.25
    
    
#    Lb=[lb]*nd
#    Ub=[ub]*nd
    convergence=[]

    # RInitialize nests randomely
    nest=numpy.random.rand(n,dim)*(ub-lb)+lb
    labelsPred = numpy.zeros((n,len(points)))
    
    new_nest=numpy.zeros((n,dim))
    new_nest=numpy.copy(nest)
    
    bestnest=[0]*dim;
    bestLabelsPred=[0]*len(points);
     
    fitness=numpy.zeros(n) 
    fitness.fill(float("inf"))
    

    s=solution()

     
    print("CS is optimizing  \""+objf.__name__+"\"")    
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    fmin,bestnest,bestlabels,nest,fitness,labelsPred =get_best_nest(nest,labelsPred,new_nest,fitness,n,dim,objf,k,points, metric)
    convergence = [];
    # Main loop counter
    for iter in range (0,N_IterTotal):
        # Generate new solutions (but keep the current best)
     
         new_nest=get_cuckoos(nest,bestnest,lb,ub,n,dim)
         
         
         # Evaluate new solutions and find best
         fnew,best,bestlabels,nest,fitness,labelsPred=get_best_nest(nest,labelsPred,new_nest,fitness,n,dim,objf,k,points, metric)
         
        
         new_nest=empty_nests(new_nest,pa,n,dim) ;
         
        
        # Evaluate new solutions and find best
         fnew,best,bestlabels,nest,fitness,labelsPred=get_best_nest(nest,labelsPred,new_nest,fitness,n,dim,objf,k,points, metric)
    
         if fnew<fmin:
            fmin=fnew
            bestnest=best
            bestLabelsPred=bestlabels
    
         if (iter%1==0):
            print(['At iteration '+ str(iter)+ ' the best fitness is '+ str(fmin)]);
         convergence.append(fmin)

    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence
    s.optimizer="CS"
    s.objfname=objf.__name__
    s.bestIndividual = bestnest
    s.labelsPred = numpy.array(bestLabelsPred, dtype=numpy.int64)
    
     
    
    return s
    




 






# -*- coding: utf-8 -*-
"""
Created on Sat Mar  9 18:12:29 2019

@author: Raneem
"""

from sklearn import cluster, metrics
from scipy.spatial.distance import pdist, cdist
import numpy
import sys

def getLabelsPred(startpts, points, k):
    labelsPred = [-1] * len(points)
    
    for i in range(len(points)):
        distances = numpy.linalg.norm(points[i]-startpts, axis = 1)
        labelsPred[i] = numpy.argmin(distances)
    
    return labelsPred
    

def SSE(startpts, points, k, metric):
    labelsPred = getLabelsPred(startpts, points, k)
    fitness = 0
        
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        centroidsForPoints = startpts[labelsPred]
        fitness = 0
        for i in range(k):
            indexes = [n for n,x in enumerate(labelsPred) if x==i]
            fit = cdist(points[indexes], centroidsForPoints[indexes], metric)**2
            fit = sum(fit)[0]
            fitness += fit
    return fitness, labelsPred


def TWCV(startpts, points, k):
    labelsPred = getLabelsPred(startpts, points, k)
    
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        sumAllFeatures = sum(sum(numpy.power(points,2)))
        sumAllPairPointsCluster = 0
        for clusterId in range(k):
            indices = numpy.where(numpy.array(labelsPred) == clusterId)[0]
            pointsInCluster = points[numpy.array(indices)]
            sumPairPointsCluster = sum(pointsInCluster)
            sumPairPointsCluster = numpy.power(sumPairPointsCluster,2)
            sumPairPointsCluster = sum(sumPairPointsCluster)
            sumPairPointsCluster = sumPairPointsCluster/len(pointsInCluster)
            
            sumAllPairPointsCluster += sumPairPointsCluster
        fitness = (sumAllFeatures - sumAllPairPointsCluster)
    return fitness, labelsPred


def SC(startpts, points, k, metric):    
    labelsPred = getLabelsPred(startpts, points, k)
    
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        silhouette = metrics.silhouette_score(points, labelsPred, metric=metric)
        #silhouette = (silhouette - (-1)) / (1 - (-1))
        silhouette = (silhouette + 1) / 2
        fitness = 1 - silhouette
    return fitness, labelsPred


def DB(startpts, points, k):
    labelsPred = getLabelsPred(startpts, points, k)
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        fitness = metrics.davies_bouldin_score(points, labelsPred)
    return fitness, labelsPred

def CH(startpts, points, k):
    labelsPred = getLabelsPred(startpts, points, k)
    
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        ch = metrics.calinski_harabaz_score(points, labelsPred)
        fitness = 1 / ch
    return fitness, labelsPred


def delta_fast(ck, cl, distances):
    values = distances[numpy.where(ck)][:, numpy.where(cl)]
    values = values[numpy.nonzero(values)]

    return numpy.min(values)
    
def big_delta_fast(ci, distances):
    values = distances[numpy.where(ci)][:, numpy.where(ci)]
    #values = values[numpy.nonzero(values)]
            
    return numpy.max(values)

def dunn_fast(points, labels, metric):
    v = pdist(points, metric)
    size_X  = len(points)
    X = numpy.zeros((size_X,size_X))
    X[numpy.triu_indices(X.shape[0], k = 1)] = v
    distances = X + X.T
    ks = numpy.sort(numpy.unique(labels))
    
    deltas = numpy.ones([len(ks), len(ks)])*1000000
    big_deltas = numpy.zeros([len(ks), 1])
    
    l_range = list(range(0, len(ks)))
    
    for k in l_range:
        for l in (l_range[0:k]+l_range[k+1:]):
            deltas[k, l] = delta_fast((labels == ks[k]), (labels == ks[l]), distances)
        
        big_deltas[k] = big_delta_fast((labels == ks[k]), distances)

    di = numpy.min(deltas)/numpy.max(big_deltas)
    return di
    

def DI(startpts, points, k, metric):
    labelsPred = getLabelsPred(startpts, points, k)
    
    if numpy.unique(labelsPred).size < k:
        fitness = sys.float_info.max
    else:
        dunn = dunn_fast(points, labelsPred, metric)
        if(dunn < 0):
            dunn = 0
        fitness = 1 - dunn
    return fitness, labelsPred


def getFunctionDetails(a):    
    # [name, lb, ub]
    param = {  0: ["SSE",0,1],
            1: ["TWCV",0,1],
            2: ["SC",0,1],
            3: ["DB",0,1],
            #4: ["CH",0,1],
            4: ["DI",0,1]
            }
    return param.get(a, "nothing")


'''
ref: https://jtemporal.com/kmeans-and-elbow-method/
ref: https://medium.com/@masarudheena/4-best-ways-to-find-optimal-number-of-clusters-for-clustering-with-python-code-706199fa957c
ref: https://github.com/minddrummer/gap/edit/master/gap/gap.py
ref: https://www.tandfonline.com/doi/pdf/10.1080/03610927408827101
ref: https://www.sciencedirect.com/science/article/pii/S0952197618300629?casa_token=W6QEUM7YA2cAAAAA:jtbAvDYF8axr8ghhr92aCnhXJ71wtCJ1tEZFHAjBUBbLrbJ8wdLHG0d4HIhDN5ICZJmrEGQ71vQ
ref: https://github.com/tirthajyoti/Machine-Learning-with-Python/blob/master/Clustering-Dimensionality-Reduction/Clustering_metrics.ipynb
'''

from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from math import sqrt
from sklearn import metrics 
import numpy as np
import random
import pandas as pd
import scipy
import sys

def min_clusters(points):
	k_list = all_methods(points)
	return min(k_list)

def max_clusters(points):
	k_list = all_methods(points)
	return max(k_list)

def median_clusters(points):
	k_list = all_methods(points)
	return int(np.median(np.array(k_list)))

def majority_clusters(points):
    k_list = all_methods(points)
    return max(set(k_list), key = k_list.count) 

def all_methods(points):
    k_list = [0] * 6
    k_list[0] = ELBOW(points)
    k_list[1] = GAP_STATISTICS(points)
    k_list[2] = SC(points)
    k_list[3] = CH(points)
    k_list[4] = DB(points)
    k_list[5] = BIC(points)
    return k_list


### Calinski-Harabasz Index
def CH(data):
	ch_max = 0
	ch_max_clusters = 2
	for n_clusters in range(2,10):
		model = KMeans(n_clusters = n_clusters)
		labels = model.fit_predict(data)
		ch_score = metrics.calinski_harabaz_score(data, labels)
		if ch_score > ch_max:
			ch_max = ch_score
			ch_max_clusters = n_clusters
	return ch_max_clusters
### END Calinski-Harabasz Index


### silhouette score 
def SC(data):
	sil_max = 0
	sil_max_clusters = 2
	for n_clusters in range(2,10):
		model = KMeans(n_clusters = n_clusters)
		labels = model.fit_predict(data)
		sil_score = metrics.silhouette_score(data, labels)
		if sil_score > sil_max:
			sil_max = sil_score
			sil_max_clusters = n_clusters
	return sil_max_clusters
### END silhouette score 

### DB score 
def DB(data):
    db_min = sys.float_info.max
    db_min_clusters = 2
    for n_clusters in range(2,10):
        model = KMeans(n_clusters = n_clusters)
        labels = model.fit_predict(data)
        db_score = metrics.davies_bouldin_score(data, labels)
        if db_score < db_min:
            db_min = db_score
            db_min_clusters = n_clusters
    return db_min_clusters
### END DB score 

### Bayesian Information Criterion
def BIC(data):
    bic_max = 0
    bic_max_clusters = 2
    for n_clusters in range(2,10):
        gm = GaussianMixture(n_components=n_clusters,n_init=10,tol=1e-3,max_iter=1000).fit(data)
        bic_score = -gm.bic(data)
        if bic_score > bic_max:
            bic_max = bic_score
            bic_max_clusters = n_clusters
    return bic_max_clusters
### END Bayesian Information Criterion


### ELBOW
def ELBOW(data):
	wcss = calculate_wcss(data)
	n_clusters = optimal_number_of_clusters(wcss)
	return n_clusters

def optimal_number_of_clusters(wcss):
    x1, y1 = 2, wcss[0]
    x2, y2 = 20, wcss[len(wcss)-1]

    distances = []
    for i in range(len(wcss)):
        x0 = i+2
        y0 = wcss[i]
        numerator = abs((y2-y1)*x0 - (x2-x1)*y0 + x2*y1 - y2*x1)
        denominator = sqrt((y2 - y1)**2 + (x2 - x1)**2)
        distances.append(numerator/denominator)
    
    return distances.index(max(distances)) + 2

def calculate_wcss(data):
    wcss = []
    for n in range(2, 10):
        kmeans = KMeans(n_clusters=n)
        kmeans.fit(X=data)
        wcss.append(kmeans.inertia_)    
    return wcss
### END ELBOW

### gap statistics
def GAP_STATISTICS(data):
	gaps, s_k, K = gap_statistic(data, refs=None, B=10, K=range(1,11), N_init = 10) 
	bestKValue = find_optimal_k(gaps, s_k, K)	
	
	return bestKValue

def short_pair_wise_D(each_cluster):
    '''
    this function computes the sum of the pairwise distance(repeatedly) of all points in one cluster;
    each pair be counting twice here; using the short formula below instead of the original meaning of pairwise distance of all points

    each_cluster: np.array, with containing all points' info within the array
    '''
    mu = each_cluster.mean(axis = 0)
    total = sum(sum((each_cluster - mu)**2)) * 2.0 * each_cluster.shape[0]
    return total

def compute_Wk(data, classfication_result):
    '''
    this function computes the Wk after each clustering

    data:np.array, containing all the data
    classfication_result: np.array, containing all the clustering results for all the data
    '''
    Wk = 0
    label_set = set(classfication_result.tolist())
    for label in label_set:
        each_cluster = data[classfication_result == label, :]
        D = short_pair_wise_D(each_cluster)
        Wk = Wk + D/(2.0*each_cluster.shape[0])
    return Wk
 
def gap_statistic(X, refs=None, B=10, K=range(2,11), N_init = 10):
    '''
    this function first generates B reference samples; for each sample, the sample size is the same as the original datasets;
    the value for each reference sample follows a uniform distribution for the range of each feature of the original datasets;
    using a simplify formula to compute the D of each cluster, and then the Wk; K should be a increment list, 1-10 is fair enough;
    the B value is about the number of replicated samples to run gap-statistics, it is recommended as 10, and it should not be changed/decreased that to a smaller value;
    
    X: np.array, the original data;
    refs: np.array or None, it is the replicated data that you want to compare with if there exists one; if no existing replicated/proper data, just use None, and the function
    will automatically generates them; 
    B: int, the number of replicated samples to run gap-statistics; it is recommended as 10, and it should not be changed/decreased that to a smaller value;
    K: list type, the range of K values to test on;
    N_init: int, states the number of initial starting points for each K-mean running under sklearn, in order to get stable clustering result each time; 
    you may not need such many starting points, so it can be reduced to a smaller number to fasten the computation;
    n_jobs below is not an argument for this function,but it clarifies the parallel computing, could fasten the computation, this can be only changed inside the script, not as an argument of the function;
    '''
    shape = X.shape
    if refs==None:
        tops = X.max(axis=0)
        bots = X.min(axis=0)
        dists = scipy.matrix(scipy.diag(tops-bots))
        rands = scipy.random.random_sample(size=(shape[0],shape[1],B))
        for i in range(B):
            rands[:,:,i] = rands[:,:,i]*dists+bots
    else:
        rands = refs

    gaps = np.zeros(len(K))
    Wks = np.zeros(len(K))
    Wkbs = np.zeros((len(K),B))

    for indk, k in enumerate(K):
        # #setup the kmean clustering instance
        #n_jobs set up for parallel:1 mean No Para-computing; -1 mean all parallel computing
        #n_init is the number of times each Kmeans running to get stable clustering results under each K value
        k_means =  KMeans(n_clusters=k, init='k-means++', n_init=N_init, max_iter=300, tol=0.0001, precompute_distances='auto', verbose=0, random_state=None, copy_x=True, n_jobs=1)
        k_means.fit(X)
        classfication_result = k_means.labels_
        #compute the Wk for the classification result
        Wks[indk] = compute_Wk(X,classfication_result)
        
        # clustering on B reference datasets for each 'k' 
        for i in range(B):
            Xb = rands[:,:,i]
            k_means.fit(Xb)
            classfication_result_b = k_means.labels_
            Wkbs[indk,i] = compute_Wk(Xb,classfication_result_b)

    #compute gaps and sk
    gaps = (np.log(Wkbs)).mean(axis = 1) - np.log(Wks)        
    sd_ks = np.std(np.log(Wkbs), axis=1)
    sk = sd_ks*np.sqrt(1+1.0/B)
    return gaps, sk, K

def find_optimal_k(gaps, s_k, K):
    '''
    this function is finding the best K value given the computed results of gap-statistics

    gaps: np.array, containing all the gap-statistics results;
    s_k: float, the baseline value to minus with; say reference paper for detailed meaning;
    K: list, containing all the tested K values;
    '''
    gaps_thres = gaps - s_k
    below_or_above = (gaps[0:-1] >= gaps_thres[1:])
    if below_or_above.any():
        optimal_k = K[below_or_above.argmax()]
    else:
        optimal_k = K[-1]
    return optimal_k

### END gap statistics


# -*- coding: utf-8 -*-
"""
Created on Thu May 26 02:00:55 2016

@author: hossam
"""
import math
import numpy
import random
import time
from .._solution import solution
    

def BAT(objf,lb,ub,dim,N,Max_iteration, k, points, metric):
    
    n=N;      # Population size
    #lb=-50
    #ub=50
    
    N_gen=Max_iteration  # Number of generations
    
    A=0.5;      # Loudness  (constant or decreasing)
    r=0.5;      # Pulse rate (constant or decreasing)
    
    Qmin=0         # Frequency minimum
    Qmax=2         # Frequency maximum
    
    
    d=dim           # Number of dimensions 
    
    # Initializing arrays
    Q=numpy.zeros(n)  # Frequency
    v=numpy.zeros((n,d))  # Velocities
    Convergence_curve=[];
    
    # Initialize the population/solutions
    Sol=numpy.random.rand(n,d)*(ub-lb)+lb
    labelsPred=numpy.zeros((n,len(points)))
    Fitness=numpy.zeros(n)

    S=numpy.zeros((n,d))
    S=numpy.copy(Sol)
    
    
    # initialize solution for the final results   
    s=solution()
    print("BAT is optimizing  \""+objf.__name__+"\"")    
    
    # Initialize timer for the experiment
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    #Evaluate initial random solutions
    for i in range(0,n):
        startpts = numpy.reshape(Sol[i,:], (k,(int)(dim/k)))
        if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
            fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
        else:
            fitnessValue, labelsPredValues=objf(startpts, points, k) 
        Fitness[i] = fitnessValue
        labelsPred[i,:] = labelsPredValues
    
    
    # Find the initial best solution
    fmin = min(Fitness)
    I=numpy.argmin(Fitness)
    best=Sol[I,:]
    bestLabelsPred=labelsPred[I,:]
       
    # Main loop
    for t in range (0,N_gen): 
        
        # Loop over all bats(solutions)
        for i in range (0,n):
          Q[i]=Qmin+(Qmin-Qmax)*random.random()
          v[i,:]=v[i,:]+(Sol[i,:]-best)*Q[i]
          S[i,:]=Sol[i,:]+v[i,:]
          
          # Check boundaries
          Sol=numpy.clip(Sol,lb,ub)

    
          # Pulse rate
          if random.random()>r:
              S[i,:]=best+0.001*numpy.random.randn(d)
          
    
          # Evaluate new solutions
          startpts = numpy.reshape(S[i,:], (k,(int)(dim/k)))

          if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
              fitnessValue, labelsPredValues=objf(startpts, points, k, metric) 
          else:
              fitnessValue, labelsPredValues=objf(startpts, points, k) 
              
          Fnew = fitnessValue
          LabelsPrednew = labelsPredValues
          
          # Update if the solution improves
          if ((Fnew != numpy.inf) and (Fnew<=Fitness[i]) and (random.random()<A) ):
                Sol[i,:]=numpy.copy(S[i,:])
                Fitness[i]=Fnew
                labelsPred[i,:]=LabelsPrednew
           
    
          # Update the current best solution
          if Fnew != numpy.inf and Fnew<=fmin:
                best=numpy.copy(S[i,:])
                fmin=Fnew
                bestLabelsPred=LabelsPrednew
                
        #update convergence curve
        Convergence_curve.append(fmin)        

        if (t%1==0):
            print(['At iteration '+ str(t)+ ' the best fitness is '+ str(fmin)])
    
    
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=Convergence_curve
    s.optimizer="BAT"   
    s.objfname=objf.__name__
    s.labelsPred = numpy.array(bestLabelsPred, dtype=numpy.int64)
    s.bestIndividual = best
    
    
    
    return s



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 15 21:04:15 2019

@author: Raneem
"""

import random
import numpy
import math
from .._solution import solution
import time

def PSO(objf,lb,ub,dim,PopSize,iters, k, points, metric):
    
    # PSO parameters
    
#    dim=30
#    iters=200
    
#    Vmax=6
    Vmax=6#raneem
#    PopSize=50     #population size
    wMax=0.9
    wMin=0.2
    c1=2
    c2=2
#    lb=-10
#    ub=10
#    
    s=solution()
    
    
    ######################## Initializations
    
    vel=numpy.zeros((PopSize,dim))
    
    pBestScore=numpy.zeros(PopSize) 
    pBestScore.fill(float("inf"))    
    pBest=numpy.zeros((PopSize,dim))
    pBestLabelsPred=numpy.full((PopSize,len(points)), numpy.inf)
    
    
    gBest=numpy.zeros(dim)
    gBestScore=float("inf")
    gBestLabelsPred=numpy.full(len(points), numpy.inf)
    
    pos=numpy.random.uniform(0,1,(PopSize,dim)) *(ub-lb)+lb
    
    convergence_curve=numpy.zeros(iters)
    
    ############################################
    print("PSO is optimizing  \""+objf.__name__+"\"")    
    
    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    for l in range(0,iters):
        for i in range(0,PopSize):
            #pos[i,:]=checkBounds(pos[i,:],lb,ub)
            pos[i,:]=numpy.clip(pos[i,:], lb, ub)
            #Calculate objective function for each particle
            
            startpts = numpy.reshape(pos[i,:], (k,(int)(dim/k)))
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitness, labelsPred=objf(startpts, points, k, metric) 
            else:
                fitness, labelsPred=objf(startpts, points, k) 
    
            if(pBestScore[i]>fitness):
                pBestScore[i]=fitness
                pBest[i,:]=pos[i,:].copy()
                pBestLabelsPred[i,:]=numpy.copy(labelsPred)
                
            if(gBestScore>fitness):
                gBestScore=fitness
                gBest=pos[i,:].copy()
                gBestLabelsPred=numpy.copy(labelsPred)
        
        #Update the W of PSO
        w=wMax-l*((wMax-wMin)/iters);#check this
        
        for i in range(0,PopSize):
            for j in range (0,dim):
                r1=random.random()
                r2=random.random()
                vel[i,j]=w*vel[i,j]+c1*r1*(pBest[i,j]-pos[i,j])+c2*r2*(gBest[j]-pos[i,j])
                
                if(vel[i,j]>Vmax):
                    vel[i,j]=Vmax
                
                if(vel[i,j]<-Vmax):
                    vel[i,j]=-Vmax
                            
                pos[i,j]=pos[i,j]+vel[i,j]
        
        convergence_curve[l]=gBestScore
        
        if (l%1==0):
               print(['At iteration '+ str(l+1)+ ' the best fitness is '+ str(gBestScore)]);
        
        
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=convergence_curve
    s.optimizer="PSO"
    s.objfname=objf.__name__
    s.labelsPred = numpy.array(gBestLabelsPred, dtype=numpy.int64)
    s.bestIndividual = gBest

    return s
         
    



import random
import numpy
import math
from .._solution import solution

from time import sleep
import time
import sys

  
def SSA(objf,lb,ub,dim,N,Max_iteration, k, points, metric):

    #Max_iteration=1000
    #lb=-100
    #ub=100
    #dim=30
    #N=50 # Number of search agents
    
    Convergence_curve=numpy.zeros(Max_iteration)

        
    #Initialize the positions of salps
    SalpPositions=numpy.random.uniform(0,1,(N,dim)) *(ub-lb)+lb
    SalpFitness=numpy.full(N,float("inf"))
    SalpLabelsPred=numpy.full((N,len(points)), numpy.inf)
    
    FoodPosition=numpy.zeros(dim)
    FoodFitness=float("inf")
    FoodLabelsPred=numpy.full(len(points), numpy.inf)
    #Moth_fitness=numpy.fell(float("inf"))
    
    s=solution()

    print("SSA is optimizing  \""+objf.__name__+"\"")    

    timerStart=time.time() 
    s.startTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    
    for i in range(0,N):
       # evaluate moths
        startpts = numpy.reshape(SalpPositions[i,:], (k,(int)(dim/k)))

        if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
            fitness, labelsPred=objf(startpts, points, k, metric) 
        else:
            fitness, labelsPred=objf(startpts, points, k) 
        SalpFitness[i] = fitness
        SalpLabelsPred[i,:] = labelsPred
        
    sorted_salps_fitness=numpy.sort(SalpFitness)
    I=numpy.argsort(SalpFitness)
    
    Sorted_salps=numpy.copy(SalpPositions[I,:])
    Sorted_LabelsPred=numpy.copy(SalpLabelsPred[I,:])
       
    
    FoodPosition=numpy.copy(Sorted_salps[0,:])
    FoodFitness=sorted_salps_fitness[0]
    FoodLabelsPred=Sorted_LabelsPred[0]
    '''
    Convergence_curve[0]=FoodFitness
    print(['At iteration 0 the best fitness is '+ str(FoodFitness)])
    '''
    Iteration=1;    
    
    # Main loop
    while (Iteration<Max_iteration):
        
        #sleep(1)
    
        # Number of flames Eq. (3.14) in the paper
        #Flame_no=round(N-Iteration*((N-1)/Max_iteration));
        
        c1 = 2*math.exp(-(4*Iteration/Max_iteration)**2); # Eq. (3.2) in the paper

        for i in range(0,N):
            
            SalpPositions= numpy.transpose(SalpPositions);

            if i<N/2:
                for j in range(0,dim):
                    c2=random.random()
                    c3=random.random()
                    #Eq. (3.1) in the paper 
                    if c3<0.5:
                        SalpPositions[j,i]=FoodPosition[j]+0.1*c1*((ub-lb)*c2+lb);
                    else:
                        SalpPositions[j,i]=FoodPosition[j]-0.1*c1*((ub-lb)*c2+lb);
                    
                    ####################
            
            
            elif i>=N/2 and i<N:
                point1=numpy.copy(SalpPositions[:,i-1]);
                point2=numpy.copy(SalpPositions[:,i]);
                
                SalpPositions[:,i]=(point2+point1)/2; # Eq. (3.4) in the paper
        
        
            SalpPositions= numpy.transpose(SalpPositions);
        
           
        for i in range(0,N):
            
        
           # Check if salps go out of the search spaceand bring it back
            SalpPositions[i,:]=numpy.clip(SalpPositions[i,:], lb, ub)            
            
            startpts = numpy.reshape(SalpPositions[i,:], (k,(int)(dim/k)))
        
            if objf.__name__ == 'SSE' or objf.__name__ == 'SC' or objf.__name__ == 'DI':
                fitness, labelsPred=objf(startpts, points, k, metric) 
            else:
                fitness, labelsPred=objf(startpts, points, k) 
                
            SalpFitness[i] = fitness
            SalpLabelsPred[i,:] = numpy.copy(labelsPred)
        
             
            if SalpFitness[i]<FoodFitness:
                FoodPosition=numpy.copy(SalpPositions[i,:])
                FoodFitness=SalpFitness[i]
                FoodLabelsPred = numpy.copy(SalpLabelsPred[i,:])
                
                
        Convergence_curve[Iteration]=FoodFitness
        #Display best fitness along the iteration
        
        
        if (Iteration%1==0):
            print(['At iteration '+ str(Iteration)+ ' the best fitness is '+ str(FoodFitness)]);
        
    
    
    
        Iteration=Iteration+1; 
    
    timerEnd=time.time()  
    s.endTime=time.strftime("%Y-%m-%d-%H-%M-%S")
    s.executionTime=timerEnd-timerStart
    s.convergence=Convergence_curve
    s.optimizer="SSA"   
    s.objfname=objf.__name__
    s.labelsPred = numpy.array(FoodLabelsPred, dtype=numpy.int64)
    s.bestIndividual = FoodPosition
    
    return s
    







# -*- coding: utf-8 -*-
"""
Created on Thu May 20 17:06:19 2021

@author: Dang Trung Anh
"""

import os
from pip._internal.network.session import PipSession
from setuptools import setup, find_packages

VERSION = '1.0.3' 
DESCRIPTION = 'An Open-Source Nature-Inspired Optimization Clustering Framework in Python'

from pip._internal.req import parse_requirements

install_reqs = parse_requirements('requirements.txt', session=PipSession())
reqs = [str(ir.requirement) for ir in install_reqs]

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
       # the name must match the folder name 'EvoCC'
        name="EvoCluster", 
        version=VERSION,
        author="Qaddoura, Raneem, Hossam Faris, Ibrahim Aljarah, and Pedro A. Castillo",
        author_email="dangtrunganh@gmail.com, raneem.qaddoura@gmail.com",
        description=DESCRIPTION,
        long_description=long_description,
        # package_dir={"": "EvoCluster"},
        packages=find_packages(),
        # add any additional packages that 
        install_requires=reqs,
        url='https://github.com/RaneemQaddoura/EvoCluster',
        keywords=['python', 'first package'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ],
        package_data = {'': ['datasets/*.csv']},
        
)


import numpy as np
import pandas as pd 
from . import _objectives
import matplotlib.pyplot as plt

def run(results_directory, optimizer, objectivefunc, dataset_List, ev_measures, Iterations):
    plt.ioff()
    
    fileResultsDetailsData = pd.read_csv(results_directory + '/experiment_details.csv')
    for d in range(len(dataset_List)):        
        dataset_filename = dataset_List[d] + '.csv' 
        for j in range (0, len(objectivefunc)):
            for z in range (0, len(ev_measures)):
                
                #Box Plot
                data = []      
                    
                for i in range(len(optimizer)): 
                    objective_name = objectivefunc[j]
                    optimizer_name = optimizer[i]
                    
                    detailedData = fileResultsDetailsData[(fileResultsDetailsData["Dataset"] == dataset_List[d]) & (fileResultsDetailsData["Optimizer"] == optimizer_name) & (fileResultsDetailsData["objfname"] == objective_name)]
                    detailedData = detailedData[ev_measures[z]]
                    detailedData = np.array(detailedData).T.tolist()
                    data.append(detailedData)

                #, notch=True
                box=plt.boxplot(data,patch_artist=True,labels=optimizer)
                

                colors = ['#5c9eb7','#f77199', '#cf81d2','#4a5e6a','#f45b18',
                '#ffbd35','#6ba5a1','#fcd1a1','#c3ffc1','#68549d',
                '#1c8c44','#a44c40','#404636']
                for patch, color in zip(box['boxes'], colors):
                    patch.set_facecolor(color)
                 
                plt.legend(handles= box['boxes'], labels=optimizer, 
                    loc="top right", bbox_to_anchor=(1.2,1.02))
                fig_name = results_directory + "/boxplot-" + dataset_List[d] + "-" + objective_name + "-" + ev_measures[z] + ".png"
                plt.savefig(fig_name, bbox_inches='tight')
                plt.clf()
                #plt.show()
                